/*
 * Copyright (c) 2002-2004, The Tendra Project <http://www.ten15.org/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 *    		 Crown Copyright (c) 1997
 *
 *    This TenDRA(r) Computer Program is subject to Copyright
 *    owned by the United Kingdom Secretary of State for Defence
 *    acting through the Defence Evaluation and Research Agency
 *    (DERA).  It is made available to Recipients with a
 *    royalty-free licence for its use, reproduction, transfer
 *    to other parties and amendment for any purpose not excluding
 *    product development provided that any such use et cetera
 *    shall be deemed to be acceptance of the following conditions:-
 *
 *        (1) Its Recipients shall ensure that this Notice is
 *        reproduced upon any copies or amended versions of it;
 *
 *        (2) Any amended version of it shall be clearly marked to
 *        show both the nature of and the organisation responsible
 *        for the relevant amendment or amendments;
 *
 *        (3) Its onward transfer from a recipient to another
 *        party shall be deemed to be that party's acceptance of
 *        these conditions;
 *
 *        (4) DERA gives no warranty or assurance as to its
 *        quality or suitability for any purpose and DERA accepts
 *        no liability whatsoever in relation to any use to which
 *        it may be put.
 *
 * $TenDRA: tendra/src/utilities/sid/scope.h,v 1.7 2004/04/01 04:51:36 bp Exp $
 */


/*** scope.h --- Scope stack ADT.
 *
 ** Author: Steve Folkes <smf@hermes.mod.uk>
 *
 *** Commentary:
 *
 * See the file "scope.c" for more information.
 *
 *** Change Log:*/

/****************************************************************************/

#ifndef H_SCOPE
#define H_SCOPE

#include "config.h"
#include "dstring.h"
#include "entry.h"
#include "table.h"

/* To avoid circularity: */
struct RuleT;

/*--------------------------------------------------------------------------*/

typedef struct ScopeMapEntryT {
	struct ScopeMapEntryT      *next;
	EntryP			from;
	EntryP			to;
} ScopeMapEntryT, *ScopeMapEntryP;

typedef struct ScopeStackFrameT {
	struct ScopeStackFrameT    *next;
	NStringT			scope;
	ScopeMapEntryP		head;
	ScopeMapEntryP	       *tail;
} ScopeStackFrameT, *ScopeStackFrameP;

typedef struct ScopeStackT {
	ScopeStackFrameP		head;
} ScopeStackT, *ScopeStackP;

/*--------------------------------------------------------------------------*/

void	scope_stack_init(ScopeStackP);
void	scope_stack_push(ScopeStackP, NStringP);
void	scope_stack_pop(ScopeStackP);
EntryP	scope_stack_add_rule(ScopeStackP, TableP, NStringP,
		struct RuleT *, BoolT *);
EntryP	scope_stack_add_action(ScopeStackP, TableP, NStringP,
		struct RuleT *, BoolT *);
EntryP	scope_stack_add_non_local(ScopeStackP, TableP, NStringP, EntryP,
		struct RuleT *);
EntryP	scope_stack_get_rule(ScopeStackP, TableP, NStringP);
EntryP	scope_stack_get_action(ScopeStackP, TableP, NStringP);
EntryP	scope_stack_get_non_local(ScopeStackP, TableP, NStringP, NStringP);
BoolT	scope_stack_check_shadowing(ScopeStackP, EntryP, struct RuleT *);

#endif /* !defined (H_SCOPE) */
